'''
Computes 3S predictions (10% schedule + 90% k-NN) and simulates their execution.
'''
import operator
import sys
import os
import re
import csv
import json
import copy

# Path of auxiliary files.
NUM_SOLVERS = 12
PORT_FILE = '/public/xcsp_instances/new/solvers/csp/fixed/portfolios_csp'
SCHED_FILE = '/public/xcsp_instances/new/predictions/csp/3S/sorted/schedule'
FEAT_PATH = '/public/xcsp_instances/new/predictions/csp/SATzilla/features/'
BEST_SOLVERS = '/public/xcsp_instances/new/solvers/csp/fixed/best_solvers'
INFO_TIMES = '/public/xcsp_instances/new/solvers/csp/fixed/infotimes_csp'
BACKUP_SOLVER = 'minisatid'
FEAT_SOLVER = 'gecode'
TIMEOUT = 1800.0
with open('/public/xcsp_instances/new/scripts/classifiers/csp/3S/rem_norm/feat_vector', 'r') as infile:
  feat_vector = json.load(infile)
with open(INFO_TIMES, 'r') as infile:
  infotimes = json.load(infile)
with open(BEST_SOLVERS, 'r') as infile:
  bs = json.load(infile)
portfolios = csv.reader(open(PORT_FILE), delimiter='|', skipinitialspace = True)

def get_schedule(test_file, num_solvers, rep, fold):
  """
  read the file containing all the schedules and pick the correct one 
  according to the tested fold    
  """
  schedule = {}
  sched_reader = csv.reader(open(SCHED_FILE, 'rb'), delimiter='|', skipinitialspace = True)
  for row in sched_reader:
    if int(row[0]) > int(num_solvers):
      break
    if int(row[0]) == int(num_solvers) and row[1] == rep and row[2] == fold: 
      schedule[row[3]] = float(row[4])
  return schedule

def compare_instances(instance1, instance2):
  """
  compute the distance between two instances
  """
  distance = 0.0
  for i in range(0, len(feat_vector[instance1])):
    d = float(feat_vector[instance1][i]) - float(feat_vector[instance2][i])
    distance += d * d
  return distance
        
def get_similar_instances(instance, training_instances, k = 10):
  """
  return the most k similar instances in the training set
  """
  distances = []
  for i in training_instances:
    distance = compare_instances(instance, i)
    distances.append((distance,i))
  distances.sort(key = lambda x : x[0])
  return distances[:k]

def get_best_solver(cluster, solvers):
  """
  Given a cluster of instances and a set of solvers, returns the best solver 
  (in solvers) according to no. of solved instances (in cluster), potentially using 
  resolution time for tie-breaking.
  """
  nsi = {}
  ast = {}
  for solver in solvers:
    nsi[solver] = 0
    ast[solver] = 0.0
  for (_,inst) in cluster:
    for solver in solvers:
      time = solver_performance(inst, solver)
      if time < TIMEOUT:
        nsi[solver] += 1
      ast[solver] += time
  maximum = max(nsi.values())
  best_solvers = filter(lambda x : nsi[x] == maximum, nsi.keys())
  sorted_solvers = sorted(best_solvers, key = lambda x : ast[x])
  return sorted_solvers[0]

def solver_performance(inst, solv):
  it = infotimes[inst][solv]
  if it[0] == "sat" or it [0] == "unsat":
    #  Here not consider features extraction time.
    return it[1]
  else:
    return TIMEOUT

def compute_solving_times(solvers, instance, training_set, schedule, solv, rep, fold):
  """
  given the list of solvers, an instance, and the list of the 
  training instances it computes the solving time of 3S
  """
  # Features extraction time.
  time = float(bs[instance]['feat'])
  #print "computing map for instance " + instance
  similar_instances = get_similar_instances(instance, training_set)
  #compute the best solver from the cluster of similar instances
  best_solver = get_best_solver(similar_instances, solvers)
  #print schedule
  solvers_map = schedule
  if best_solver in solvers_map:
    del solvers_map[best_solver]
  schedule = solvers_map.items()
  schedule.append((best_solver, TIMEOUT))
  schedule.append((BACKUP_SOLVER, TIMEOUT))
  #print schedule
  # Simulate portfolio execution and print the results.
  for (solver, timeout) in schedule:
    perf_time = solver_performance(instance, solver)
    # Not reconvert in FlatZinc if already done.
    if solver == FEAT_SOLVER:
      perf_time -= float(bs[instance]['mzn2fzn'])
    if timeout >= perf_time:
      time += perf_time
      if time < TIMEOUT:
        print instance + "|1|" + str(time) + "|"
      else:
        print instance + "|0|" + str(TIMEOUT) + "|"
      return
    else:
      time += timeout
  print instance + "|0|" + str(TIMEOUT) + "|"

def compute_predictions(solv, rep, fold, portfolio):
  '''
  Compute 3S predictions.
  '''
  # compute the instances of the train/test set.
  train_file = FEAT_PATH + 'train_rep' + rep + '_fold' + fold + '.features'
  test_file = FEAT_PATH + 'test_rep' + rep + '_fold' + fold + '.features'
  train_reader = csv.reader(open(train_file, "rb"), delimiter=' ', skipinitialspace = True)
  test_reader = csv.reader(open(test_file, "rb"), delimiter=' ', skipinitialspace = True)
  train_set = []
  test_set = []
  for row in train_reader:
    train_set.append(row[0])
  for row in test_reader:
    test_set.append(row[0])
  #get the static schedule
  schedule = get_schedule(test_file, solv, rep, fold)
  for i in test_set:
    compute_solving_times(portfolio, i, train_set, copy.copy(schedule), solv, rep, fold)

def main(args):
  n_solver = args[0]
  rep      = args[1]
  fold     = args[2]
  # Look for the portfolio of size SOLV.
  for row in portfolios:
    if row[0] == n_solver:
      portfolio = eval(row[1])
      break
  compute_predictions(n_solver, rep, fold, portfolio)

if __name__ == '__main__':
    main(sys.argv[1:])
