"""
Compute SUNNY schedules and simulates their execution.
"""
import sys
import csv
import json
from combinations import *
from math import sqrt

# Path of auxiliary files.
PORT_FILE = '/public/xcsp_instances/new/predictions/csp/3S/sorted/portfolios_csp'
BEST_SOLVERS = '/public/xcsp_instances/new/solvers/csp/fixed/best_solvers'
INFO_TIMES = '/public/xcsp_instances/new/solvers/csp/fixed/infotimes_csp'
BACKUP_SOLVER = 'minisatid'
FEAT_SOLVER = 'gecode'
TIMEOUT = 1800.0
DATASET = set(open('/public/xcsp_instances/new/dataset/csp/csp_final'))
TEST_PATH = '/public/xcsp_instances/new/dataset/csp/train_test/test_'
FEAT_VECTOR = '/public/xcsp_instances/new/scripts/classifiers/csp/3S/rem_norm/feat_vector'
with open(FEAT_VECTOR, 'r') as infile:
  feat_vector = json.load(infile)
with open(INFO_TIMES, 'r') as infile:
  infotimes = json.load(infile)
with open(BEST_SOLVERS, 'r') as infile:
  bs = json.load(infile)
portfolios = csv.reader(open(PORT_FILE), delimiter='|', skipinitialspace = True)

def compare_instances(instance1, instance2):
  """
  Compute the Euclidean distance between two instances.
  """
  distance = 0.0
  for i in range(0, len(feat_vector[instance1])):
    d = float(feat_vector[instance1][i]) - float(feat_vector[instance2][i])
    distance += d * d
  return sqrt(distance)
        
def get_similar_instances(instance, training_instances, k = 10):
  """
  Return the most k similar instances in the training set.
  """
  distances = []
  for i in training_instances:
    distance = compare_instances(instance, i)
    distances.append((distance,i))
  distances.sort(key = lambda x : x[0])
  return distances[:k]

def get_schedule(neighborhood):
  """
  Given the neighborhood of a given instance, return the corresponding 
  SUNNY schedule.
  """
  solved = {}
  times  = {}
  for solver in PORTFOLIO:
    solved[solver] = set([])
    times[solver] = 0.0
  for (_,inst) in neighborhood:
    for solver in PORTFOLIO:
      time = solver_performance(inst, solver)
      if time < TIMEOUT:
        solved[solver].add(inst)
      times[solver] += time
  max_solved = 0
  min_time = 1000000
  best_portfolio = []
  # Select the best sub-portfolio.
  for i in range(0, SOLV + 1):
    for j in range(0, binom(SOLV, i)):
      solved_instances = set([])
      solving_time = 0
      sub_portfolio = get_subset(j, i, PORTFOLIO)
      for solver in sub_portfolio:
        solved_instances.update(solved[solver])
        solving_time += times[solver]
      num_solved = len(solved_instances)
      if num_solved >  max_solved or \
        (num_solved == max_solved and solving_time < min_time):
          min_time = solving_time
          max_solved = num_solved
          best_portfolio = sub_portfolio
  # n is the number of instances solved by each solver plus the instances 
  # that no solver can solver.
  n = sum([len(solved[s]) for s in best_portfolio]) + (K - max_solved)
  schedule = {}
  # Compute the schedule and sort it by number of solved instances.
  for solver in best_portfolio:
    ns = len(solved[solver])
    if ns == 0:
      continue
    schedule[solver] = TIMEOUT / n * ns
  tot_time = sum(schedule.values())
  # Allocate to the backup solver the (eventual) remaining time.
  if tot_time < TIMEOUT:
    if BACKUP_SOLVER in schedule.keys():
      schedule[BACKUP_SOLVER] += TIMEOUT - tot_time
    else:
      schedule[BACKUP_SOLVER]  = TIMEOUT - tot_time
  sorted_schedule = sorted(
    schedule.items(), 
     key = lambda x: len(solved[x[0]]), 
     reverse = True
  )
  #print sorted_schedule
  assert(sum(t for (s,t) in sorted_schedule) <= TIMEOUT)
  return sorted_schedule

def solver_performance(inst, solv):
  """
  Return the runtime of solver solv on instance inst.
  """
  it = infotimes[inst][solv]
  if it[0] == "sat" or it [0] == "unsat":
    # Here not consider the features extraction time.
    return it[1]
  else:
    return TIMEOUT

def compute_inst_schedule(instance, training_set):
  """
  Given an instance and the training set, compute the corresponding schedule, 
  simulate it and print the overcome on std output.
  """
  # Features extraction time.
  time = float(bs[instance]['feat'])
  similar_instances = get_similar_instances(instance, training_set, K)
  schedule = get_schedule(similar_instances)
  # Simulate portfolio execution and print the results.
  for (solver, timeout) in schedule:
    perf_time = solver_performance(instance, solver)
    # Not reconvert in FlatZinc if already done.
    if solver == FEAT_SOLVER:
      perf_time -= float(bs[instance]['mzn2fzn'])
    if timeout >= perf_time:
      time += perf_time
      if time < TIMEOUT:
        print instance + "|1|" + str(time) + "|" + solver + "|" + str(schedule) + "|"
      else:
        print instance + "|0|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"
      return
    else:
      time += timeout
  print instance + "|0|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"

def compute_predictions():
  """
  Predict a schedule of solvers for each instance of the test set.
  """
  # compute the instances of the train/test set.
  test_set = set(open(TEST_PATH + 'rep' + REP + '_fold' + FOLD))
  train_set = list(DATASET - test_set)
  train_set = [inst[0 : len(inst) - 1] for inst in train_set]
  #print len(test_set), len(train_set)
  for inst in test_set:
    compute_inst_schedule(inst[0 : len(inst) - 1], train_set)

def main(args):
  """
  Given the number of solvers, repetition, fold and the k parameter of k-NN 
  print on std output the overcome of SUNNY simulation for every instance of 
  the test set.
  """
  global SOLV, REP, FOLD, K, PORTFOLIO
  SOLV  = args[0]
  REP   = args[1]
  FOLD  = args[2]
  K     = int(args[3])
  # Look for the portfolio of size SOLV.
  for row in portfolios:
    if row[0] == SOLV:
      PORTFOLIO = eval(row[1])
      break
  SOLV = int(SOLV)
  compute_predictions()

if __name__ == '__main__':
    main(sys.argv[1:])
